#include "port.h"
#include "apu.h"
#include "soundux.h"

#ifdef GP2XDEBUG
#include <stdio.h>
#include <stdlib.h>
#endif

START_EXTERN_C

struct SAPU APU;
struct SIAPU IAPU;
struct SAPURegisters APURegisters;
SSoundData SoundData;

END_EXTERN_C

volatile SoundStatus so;

int Echo [24000];
int DummyEchoBuffer [SOUND_BUFFER_SIZE];
int MixBuffer [SOUND_BUFFER_SIZE];
int EchoBuffer [SOUND_BUFFER_SIZE];
int FilterTaps [8];
unsigned long Z = 0;
int Loop [16];

uint16 SignExtend [2] = {
    0x00, 0xff00
};

START_EXTERN_C
int NoiseFreq [32] = {
    0, 16, 21, 25, 31, 42, 50, 63, 84, 100, 125, 167, 200, 250, 333,
    400, 500, 667, 800, 1000, 1300, 1600, 2000, 2700, 3200, 4000,
    5300, 6400, 8000, 10700, 16000, 32000
};

END_EXTERN_C

#define RED 0xF800
#define GREEN 0x07E0
#define BLUE 0x001F

#ifndef GP2XDEBUG
volatile uint8 *mixbuf = (volatile uint8 *)0x00200000;
volatile uint16 *gpioouth = (volatile uint16 *)0xC000106E;
volatile uint16 *rtcstcnt = (volatile uint16 *)0xC0000C08;

uint16 last, tick, cnt, led, i, inports, addr, val;
int mixcnt, soundcnt;
#else
uint8 *mixbuf[48000];
#endif

#ifdef APUDEBUG
void aputrace(uint32 val)
{
	static volatile uint32 *trace = (volatile uint32 *)0x00100000;

    if((int)trace < (int)mixbuf) {
        *trace = val;
        trace++;
    }
}
#endif

uint8 S9xAPUGetByte (uint32 Address);
void APU_EXECUTE1();
void S9xMixSamples (uint8 *buffer, int sample_count);

int main()
{
    S9xResetAPU();
	S9xInitSound(0, 1, 0);
	S9xSetPlaybackRate(24000);

	last = *rtcstcnt;
	led = 0;
	inports = 0;

    while(1) {
        //APU.OutPorts[12] = APURegisters.PC;
        //APU.OutPorts[13] = S9xAPUGetByte(APU.OutPorts[12]);
		for(i=100;i;i++)
		{
			APU_EXECUTE1();

			if(APU.InPorts[10] != inports)
			{
				addr = APU.InPorts[inports];
				val = addr & 0xff;
				addr >>= 8;
				IAPU.RAM[addr] = val;
				inports = (inports + 1) & 7;
			}
		}

		if((tick = *rtcstcnt) != last)
		{
			if(tick < last)
				cnt = tick + 1024 - last;
			else
				cnt = tick - last;
			last = tick;
			soundcnt += cnt;

			/*
			if((tick < 512) && (!led)) {
				*gpioouth |= (1 << 4);
				led = 1;
			} else if((tick >= 512) && led) {
				*gpioouth &= ~(1 << 4);
				led = 0;
			}
			*/

			if (APU.TimerEnabled [0])
			{
				APU.Timer [0] += cnt << 1;
				while (APU.Timer [0] >= APU.TimerTarget [0])
				{
					IAPU.RAM [0xfd] = (IAPU.RAM [0xfd] + 1) & 0xf;
					APU.Timer [0] -= APU.TimerTarget [0];
				}
			}
			if (APU.TimerEnabled [1])
			{
				APU.Timer [1] += cnt << 1;
				while (APU.Timer [1] >= APU.TimerTarget [1])
				{
					IAPU.RAM [0xfe] = (IAPU.RAM [0xfe] + 1) & 0xf;
					APU.Timer [1] -= APU.TimerTarget [1];
				}
			}
            if (APU.TimerEnabled [2])
            {
                APU.Timer [2] += cnt << 4;
                while (APU.Timer [2] >= APU.TimerTarget [2])
                {
                    IAPU.RAM [0xff] = (IAPU.RAM [0xff] + 1) & 0xf;
                    APU.Timer [2] -= APU.TimerTarget [2];
                }
            }

			if(soundcnt > 15) {
				S9xMixSamples((uint8 *)mixbuf + (mixcnt << 1), 750);
				if(mixcnt < 48000)
					mixcnt += 750;
				else
					mixcnt = 750;

				*gpioouth ^= 1 << 4;
				APU.OutPorts[10] = mixcnt;
				soundcnt -= 16;
			}
			/*
			if(APU.InPorts[11]) {
				S9xMixSamples((uint8 *)mixbuf, 256);
				APU.InPorts[11] -= 256;
			}
			*/
		}
    }
    return 0;
}
