#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <SDL.h>

#include "apu.h"
#include "soundux.h"

/* GP2X button mapping */
enum MAP_KEY
{
	VK_UP         , // 0
	VK_UP_LEFT    , // 1
	VK_LEFT       , // 2
	VK_DOWN_LEFT  , // 3
	VK_DOWN       , // 4
	VK_DOWN_RIGHT , // 5
	VK_RIGHT      , // 6
	VK_UP_RIGHT   , // 7
	VK_START      , // 8
	VK_SELECT     , // 9
	VK_FL         , // 10
	VK_FR         , // 11
	VK_FA         , // 12
	VK_FB         , // 13
	VK_FX         , // 14
	VK_FY         , // 15
	VK_VOL_UP     , // 16
	VK_VOL_DOWN   , // 17
	VK_TAT          // 18
};

/* The screen surface, joystick device */
SDL_Surface *screen = NULL;
SDL_Joystick *joy = NULL;

void Terminate(void)
{
	SDL_Quit();
#ifdef GP2X
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif
}

struct SAPU APU;
struct SIAPU IAPU;
struct SAPURegisters APURegisters;
SSoundData SoundData;
volatile SoundStatus so;

int Echo [24000];
int DummyEchoBuffer [SOUND_BUFFER_SIZE];
int MixBuffer [SOUND_BUFFER_SIZE];
int EchoBuffer [SOUND_BUFFER_SIZE];
int FilterTaps [8];
unsigned long Z = 0;
int Loop [16];

uint16 SignExtend [2] = {
    0x00, 0xff00
};

START_EXTERN_C
int NoiseFreq [32] = {
    0, 16, 21, 25, 31, 42, 50, 63, 84, 100, 125, 167, 200, 250, 333,
    400, 500, 667, 800, 1000, 1300, 1600, 2000, 2700, 3200, 4000,
    5300, 6400, 8000, 10700, 16000, 32000
};

#define RED 0xF800
#define GREEN 0x07E0
#define BLUE 0x001F

uint8 *mixbuf[48000];

void APU_EXECUTE1();
void S9xMixSamples (uint8 *buffer, int sample_count);
int last, next, now, cnt, col, color;
uint32 lasttcnt, nexttcnt;

int main (int argc, char *argv[])
{
	int done;

	/* Initialize SDL */
	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit (1);
	}
	atexit (Terminate);

	SDL_ShowCursor(SDL_DISABLE);

	/* Set 320x240 16-bits video mode */
	screen = SDL_SetVideoMode (320, 240, 16, SDL_SWSURFACE);
	if (screen == NULL) {
		fprintf (stderr, "Couldn't set 320x240x16 video mode: %s\n", SDL_GetError ());
		exit (2);
	}

	/* Check and open joystick device */
	if (SDL_NumJoysticks() > 0) {
		joy = SDL_JoystickOpen(0);
		if(!joy) {
			fprintf (stderr, "Couldn't open joystick 0: %s\n", SDL_GetError ());
		}
	}

#ifdef GP2X
	/* Only use GP2X code here */
#endif

#ifdef WIN32
	/* Only use Windows code here */
#endif

    S9xInitAPU();
    S9xResetAPU();

	done = 0;
	while (!done)
	{
		SDL_Event event;

		/* Check for events */
		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					/* if press Ctrl + C, terminate program */
					if ( (event.key.keysym.sym == SDLK_c) && (event.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL)) )
						done = 1;
					break;
				case SDL_KEYUP:
					break;
				case SDL_JOYBUTTONDOWN:
					/* if press Start button, terminate program */
					if ( event.jbutton.button == VK_START )
						done = 1;
					break;
				case SDL_JOYBUTTONUP:
					break;
				case SDL_QUIT:
					done = 1;
					break;
				default:
					break;
			}
		}

		/* Processing */
        printf("PC: %04x(%02x), W: %04x, P: %02x, X: %02x, S: %02x\n", APURegisters.PC, IAPU.RAM[APURegisters.PC], APURegisters.YA.W, APURegisters.P, APURegisters.X, APURegisters.S);
        APU_EXECUTE1();

        if(APU.InPorts[4])
        {
            APU.InPorts[4] = 0;
            //ptrs[2] = APU.Cycles;
            APU.Cycles = 0;
            //*gpioouth ^= (1 << 4);
        }

        if(APU.InPorts[6])
        {
            APU.InPorts[6] = 0;
            //*gpioouth ^= (1 << 4);
            if (APU.TimerEnabled [0])
            {
                APU.Timer [0] = APU.InPorts[7];
                while (APU.Timer [0] >= APU.TimerTarget [0])
                {
                    IAPU.RAM [0xfd] = (IAPU.RAM [0xfd] + 1) & 0xf;
                    APU.Timer [0] -= APU.TimerTarget [0];
                }
            }
            if (APU.TimerEnabled [1])
            {
                APU.Timer [1] = APU.InPorts[8];
                while (APU.Timer [1] >= APU.TimerTarget [1])
                {
                    IAPU.RAM [0xfe] = (IAPU.RAM [0xfe] + 1) & 0xf;
                    APU.Timer [1] -= APU.TimerTarget [1];
                }
            }
            if (APU.TimerEnabled [2])
            {
                APU.Timer [2] = APU.InPorts[9];
                while (APU.Timer [2] >= APU.TimerTarget [2])
                {
                    IAPU.RAM [0xff] = (IAPU.RAM [0xff] + 1) & 0xf;
                    APU.Timer [2] -= APU.TimerTarget [2];
                }
            }
            IAPU.RAM[0xf4] = APU.InPorts[0];
            IAPU.RAM[0xf5] = APU.InPorts[1];
            IAPU.RAM[0xf6] = APU.InPorts[2];
            IAPU.RAM[0xf7] = APU.InPorts[3];
        }
/*
        if(now >= next) {
            S9xMixSamples((uint8 *)mixbuf + (*mixcnt << 1), 750);
            if(*mixcnt < 48000)
                *mixcnt += 750;
            else
                *mixcnt = 750;
            next += 8;
        }
*/
	}

	return 0;
}
